import logging
import threading
import hashlib
from functools import lru_cache

import graphsignal
from graphsignal.sketches.kll import KLLSketch
from graphsignal import statistics
from graphsignal import metrics_pb2

logger = logging.getLogger('graphsignal')

MAX_EXCEPTIONS = 10


class PredictionRecord(object):
    __slots__ = [
        'features',
        'feature_names',
        'outputs',
        'output_names',
        'timestamp'
    ]

    def __init__(
            self,
            features=None,
            feature_names=None,
            outputs=None,
            output_names=None,
            timestamp=None):
        self.features = features
        self.feature_names = feature_names
        self.outputs = outputs
        self.output_names = output_names
        self.timestamp = timestamp


class ExceptionRecord(object):
    __slots__ = [
        'message',
        'extra_info',
        'stack_trace',
        'timestamp'
    ]

    def __init__(
            self,
            message=None,
            extra_info=None,
            stack_trace=None,
            timestamp=None):
        self.message = message
        self.extra_info = extra_info
        self.stack_trace = stack_trace
        self.timestamp = timestamp


class GroundTruthRecord(object):
    __slots__ = [
        'label',
        'prediction',
        'prediction_timestamp',
        'segments'
    ]

    def __init__(
            self,
            label=None,
            prediction=None,
            prediction_timestamp=None,
            segments=None):
        self.label = label
        self.prediction = prediction
        self.prediction_timestamp = prediction_timestamp
        self.segments = segments


class WindowUpdater(object):
    __slots__ = [
        '_window_proto',
        '_update_lock',
        '_is_empty',
        '_metric_updaters',
        '_prediction_records',
        '_prediction_batch_size',
        '_ground_truth_records',
    ]

    def __init__(self):
        self._window_proto = metrics_pb2.PredictionWindow()
        self._update_lock = threading.Lock()
        self._is_empty = True
        self._metric_updaters = {}
        self._prediction_records = []
        self._prediction_batch_size = 0
        self._ground_truth_records = []

    def window(self):
        return self._window_proto

    def is_empty(self):
        return self._is_empty

    def add_prediction(self, prediction, batch_size):
        with self._update_lock:
            self._prediction_records.append(prediction)
            self._prediction_batch_size += batch_size
            self._is_empty = False

        if self._prediction_batch_size >= graphsignal._get_config().buffer_size:
            self._update_predictions()

    def add_exception(self, exception):
        with self._update_lock:
            self._window_proto.num_exceptions += 1

            if len(self._window_proto.exceptions) < MAX_EXCEPTIONS:
                exception_proto = self._window_proto.exceptions.add()
                exception_proto.message = exception.message
                if exception.extra_info is not None:
                    for name, value in exception.extra_info.items():
                        exception_proto.extra_info[name] = value
                if exception.stack_trace is not None:
                    exception_proto.stack_trace = exception.stack_trace
                exception_proto.create_ts = exception.timestamp
            self._is_empty = False

    def add_ground_truth(self, ground_truth):
        with self._update_lock:
            self._ground_truth_records.append(ground_truth)
            self._is_empty = False

        if len(self._ground_truth_records) >= graphsignal._get_config().buffer_size:
            self._update_ground_truth()

    def _update_predictions(self):
        if len(self._prediction_records) > 0:
            with self._update_lock:
                if self._window_proto.start_ts == 0:
                    self._window_proto.start_ts = self._prediction_records[0].timestamp
                else:
                    self._window_proto.start_ts = min(
                        self._window_proto.start_ts, self._prediction_records[0].timestamp)
                self._window_proto.end_ts = max(
                    self._window_proto.end_ts, self._prediction_records[-1].timestamp)
                self._window_proto.num_predictions += self._prediction_batch_size

                try:
                    statistics.update_data_metrics(
                        self._metric_updaters, self._window_proto, self._prediction_records)
                except BaseException:
                    logger.error('Error updating data metrics', exc_info=True)

                self._prediction_batch_size = 0
                self._prediction_records = []

    def _update_ground_truth(self):
        if len(self._ground_truth_records) > 0:
            with self._update_lock:
                if self._window_proto.start_ts == 0:
                    self._window_proto.start_ts = self._ground_truth_records[0].prediction_timestamp
                else:
                    self._window_proto.start_ts = min(
                        self._window_proto.start_ts, self._ground_truth_records[0].prediction_timestamp)
                self._window_proto.end_ts = max(
                    self._window_proto.end_ts, self._ground_truth_records[-1].prediction_timestamp)
                self._window_proto.num_ground_truths += len(
                    self._ground_truth_records)

                try:
                    statistics.update_performance_metrics(
                        self._metric_updaters, self._window_proto, self._ground_truth_records)
                except BaseException:
                    logger.error('Error updating data metrics', exc_info=True)

                self._ground_truth_records = []

    def finalize(self):
        self._update_predictions()
        self._update_ground_truth()

        with self._update_lock:
            for metric_updater in self._metric_updaters.values():
                metric_updater.finalize()


class MetricUpdater(object):
    __slots__ = [
        '_metric_proto',
        '_sketch'
    ]

    def __init__(self, metric_proto, name, dimensions=None):
        self._metric_proto = metric_proto
        self._sketch = None

        self._metric_proto.name = name
        if dimensions is not None:
            for name, value in dimensions.items():
                self._metric_proto.dimensions[name] = value

    def update_gauge(self, value):
        if self._metric_proto.type == self._metric_proto.NOT_INITIALIZED:
            self._metric_proto.type = self._metric_proto.GAUGE
        self._metric_proto.gauge_value.gauge = value

    def update_counter(self, value):
        if self._metric_proto.type == self._metric_proto.NOT_INITIALIZED:
            self._metric_proto.type = self._metric_proto.COUNTER
            self._metric_proto.counter_value.counter = 0
        self._metric_proto.counter_value.counter += value

    def update_ratio(self, value, total):
        if self._metric_proto.type == self._metric_proto.NOT_INITIALIZED:
            self._metric_proto.type = self._metric_proto.RATIO
            self._metric_proto.ratio_value.counter = 0
            self._metric_proto.ratio_value.total = 0
        self._metric_proto.ratio_value.counter += value
        self._metric_proto.ratio_value.total += total

    def update_distribution(self, values):
        if len(values) == 0:
            return

        if self._metric_proto.type == self._metric_proto.NOT_INITIALIZED:
            self._metric_proto.type = self._metric_proto.DISTRIBUTION
            self._metric_proto.distribution_value.sketch_impl = self._metric_proto.distribution_value.KLL10
            k_val = 128 if isinstance(values[0], (int, float)) else 10
            self._sketch = KLLSketch(k=k_val)

        for value in values:
            self._sketch.update(value)

    def finalize(self):
        if self._metric_proto.type == self._metric_proto.DISTRIBUTION:
            self._sketch.to_proto(
                self._metric_proto.distribution_value.sketch_kll10)


def get_data_stream(window_proto, data_source):
    data_stream = window_proto.data_streams[str(data_source)]
    if data_stream.data_source == data_stream.DataSource.NOT_INITIALIZED:
        data_stream.data_source = data_source
    return data_stream


def get_metric_updater(
        metric_updaters, data_stream_proto, name, dimensions=None):
    metric_key = '{0}:{1}:{2}'.format(
        data_stream_proto.data_source,
        name,
        str(dimensions) if dimensions is not None else '')

    if metric_key not in metric_updaters:
        metric_id = _sha1('{0}:{1}:{2}'.format(
            data_stream_proto.data_source,
            name,
            canonical_string(dimensions) if dimensions is not None else ''),
            size=12)
        metric = metric_updaters[metric_key] = MetricUpdater(
            data_stream_proto.metrics[metric_id], name, dimensions)
        return metric
    else:
        return metric_updaters[metric_key]


def canonical_string(obj):
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, list):
        return ','.join([str(v) for v in obj])
    elif isinstance(obj, dict):
        return ','.join(['{0}={1}'.format(k, v)
                        for k, v in dict(sorted(obj.items())).items()])
    else:
        raise ValueError('Type not supported')


@lru_cache(maxsize=2500)
def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]



import unittest
import logging
import time

import graphsignal
from graphsignal.sketches.kll import KLLSketch
from graphsignal import metrics_pb2

logger = logging.getLogger('graphsignal')


class KLLSketchTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_kll_update(self):
        k1 = KLLSketch()
        k1.update(1)
        k1.update(2)
        k1.update(2)
        k1.update(3.333)
        k1.update(1000)
        k1.update(0.0001)
        self.assertEqual(k1.count(), 6)
        self.assertEqual(
            k1.ranks(), [
                (0.0001, 1), (1, 2), (2, 3), (2, 4), (3.333, 5), (1000, 6)])
        self.assertEqual(k1.cdf(), [(0.0001, 0.16666666666666666), (1, 0.3333333333333333), (
            2, 0.5), (2, 0.6666666666666666), (3.333, 0.8333333333333334), (1000, 1.0)])

    def test_kll_update_str(self):
        k1 = KLLSketch()
        k1.update('1')
        k1.update('2')
        k1.update('2')
        k1.update('3')
        self.assertEqual(k1.count(), 4)
        self.assertEqual(k1.ranks(), [('1', 1), ('2', 2), ('2', 3), ('3', 4)])
        self.assertEqual(
            k1.cdf(), [
                ('1', 0.25), ('2', 0.5), ('2', 0.75), ('3', 1.0)])

    def test_kll_merge(self):
        k1 = KLLSketch()

        k2 = KLLSketch()
        k2.update(1)
        k2.update(2)
        k1.merge(k2)

        k3 = KLLSketch()
        k3.update(1)
        k3.update(2)
        k3.update(12)
        k1.merge(k3)

        self.assertEqual(
            k1.cdf(), [
                (1, 0.2), (1, 0.4), (2, 0.6), (2, 0.8), (12, 1.0)])

    def test_kll_distribution(self):
        k1 = KLLSketch()
        for i in range(5):
            k1.update(1)
        k1.update(2)
        for i in range(100):
            k1.update(4)

        self.assertEqual(k1.distribution(), [[1, 5], [2, 1], [4, 100]])

    def test_kll_proto(self):
        k1 = KLLSketch()
        k1.update(1)
        k1.update(2)
        k1.update(3)

        window = metrics_pb2.PredictionWindow()
        proto = window.data_streams['1'].metrics['1'].distribution_value.sketch_kll10

        k1.to_proto(proto)

        self.assertEqual(proto.item_type, proto.ItemType.DOUBLE)

        k2 = KLLSketch()
        k2.from_proto(proto)

        self.assertEqual(k2._k, k1._k)
        self.assertEqual(k2._c, k1._c)
        self.assertEqual(k2._H, k1._H)
        self.assertEqual(k2._size, k1._size)
        self.assertEqual(k2._max_size, k1._max_size)
        self.assertEqual(k2.count(), k1.count())
        self.assertEqual(k2.ranks(), k1.ranks())
        self.assertEqual(k2.cdf(), k1.cdf())

    def test_kll_update_perf(self):
        s = KLLSketch()

        start = time.time()

        #import cProfile
        #from pstats import Stats, SortKey

        # with cProfile.Profile() as pr:
        for i in range(1000):
            s.update(i)

        #stats = Stats(pr)
        # stats.sort_stats(SortKey.CUMULATIVE).print_stats(25)

        took = time.time() - start
        print('KLL update (1000) took: ', took)
        #print('CDF', s.cdf())
        self.assertTrue(took < 1)

    def test_kll_merge_perf(self):
        import random
        sketches = []
        for i in range(1000):
            s = KLLSketch()
            s.update(random.randint(1, 100))
            sketches.append(s)

        start = time.time()

        ns = KLLSketch()
        for s in sketches:
            ns.merge(s)

        took = time.time() - start
        print('kll merge (1000) took: ', time.time() - start)
        #print('kll size: ', len(ns.serialize()))
        self.assertTrue(took < 1)



import unittest
import logging
import time
from unittest.mock import patch, Mock
import sys
import pprint
import numpy as np

import graphsignal
from graphsignal.uploader import Uploader
from graphsignal import metrics_pb2

logger = logging.getLogger('graphsignal')


class SessionsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.sessions.reset_all()
        graphsignal.configure(api_key='k1', debug_mode=True)

    def tearDown(self):
        graphsignal.sessions.reset_all()
        graphsignal.shutdown()

    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    def test_session(self, mocked_upload_window, mocked_flush):
        now = int(time.time())
        session = graphsignal.session('d1')
        session.log_metadata(key='k1', value='v1')

        session.log_prediction(
            features={'A': 1, 'B': 2},
            output=False,
            actual_timestamp=now)

        session.log_prediction(
            features={'A': 10, 'B': 20},
            output=True,
            actual_timestamp=now)

        session.log_ground_truth(
            label='c1',
            prediction='c1',
            prediction_timestamp=now,
            segments=[
                's1',
                's2'])
        session.log_ground_truth(
            label='c1',
            prediction='c2',
            prediction_timestamp=now,
            segments=[
                's1',
                's3'])

        try:
            raise Exception('ex1')
        except Exception as ex:
            session.log_exception(
                message=ex, extra_info={
                    'k1': 'v1'}, exc_info=True, actual_timestamp=now)

        graphsignal.sessions.upload_all(force=True)

        mocked_upload_window.assert_called_once()

        uploaded_window = mocked_upload_window.call_args[0][0]
        self.assertEqual(uploaded_window.num_predictions, 2)
        self.assertEqual(uploaded_window.num_ground_truths, 2)
        self.assertEqual(uploaded_window.model.metadata['k1'], 'v1')
        self.assertEqual(
            uploaded_window.exceptions[0].message,
            'Exception: ex1')
        self.assertEqual(
            uploaded_window.exceptions[0].extra_info, {
                'k1': 'v1'})
        self.assertEqual(uploaded_window.num_exceptions, 1)
        self.assertTrue(uploaded_window.exceptions[0].stack_trace)
        self.assertEqual(len(uploaded_window.data_streams), 3)

    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    def test_prediction_not_uploaded(self, mocked_upload_window, mocked_flush):
        session = graphsignal.session(deployment_name='d1')

        session.log_prediction_batch(features=[[1, 2], [3, 4]])
        session.log_prediction_batch(features=[[1, 2], [3, 4]])
        session._current_window_updater._update_predictions()

        self.assertEqual(
            session._current_window_updater._window_proto.num_predictions, 4)
        mocked_upload_window.assert_not_called()

    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    def test_prediction_uploaded(
            self, mocked_upload_window, mocked_flush):
        session = graphsignal.session('d1')

        session.log_prediction_batch(
            features=[[1, 2], [3, 4]], actual_timestamp=int(time.time()) - 400)
        session.log_prediction_batch(
            features=[[1, 2], [3, 4]], actual_timestamp=int(time.time()))

        mocked_upload_window.assert_called_once()



import time
import logging
import threading
import sys
import traceback
import platform
from google.protobuf.json_format import MessageToDict

import graphsignal
from graphsignal import statistics
from graphsignal.windows import PredictionRecord, ExceptionRecord, GroundTruthRecord, WindowUpdater
from graphsignal import metrics_pb2

logger = logging.getLogger('graphsignal')

MAX_METADATA_SIZE = 10
MAX_EXTRA_INFO_SIZE = 10
MAX_SEGMENT_LENGTH = 50

_session_index = {}
_session_index_lock = threading.Lock()


class Session(object):
    __slots__ = [
        '_deployment_name',
        '_metadata',
        '_update_lock',
        '_current_window_updater',
        '_current_window_timestamp'
    ]

    def __init__(self, deployment_name):
        self._deployment_name = deployment_name
        self._metadata = {}
        self._update_lock = threading.Lock()
        self._current_window_updater = None
        self._current_window_timestamp = None

    def _tumble_window(self, timestamp, force=False, flush=True):
        window_seconds = graphsignal._get_config().window_seconds
        time_bucket = int(timestamp / window_seconds) * window_seconds

        if not force and time_bucket == self._current_window_timestamp:
            # window matches, no need to tumble/create
            return True

        if self._current_window_timestamp and time_bucket < self._current_window_timestamp:
            logger.error(
                'Cannot tumble time windows backwards. Please log in chronological order when providing timestamps explicitely.')
            return False

        if self._current_window_updater and not self._current_window_updater.is_empty():
            with self._update_lock:
                self._current_window_updater.finalize()
                window = self._current_window_updater.window()
                self._current_window_updater = WindowUpdater()
                self._current_window_timestamp = time_bucket

            window.model.deployment_name = self._deployment_name
            for key, value in self._metadata.items():
                window.model.metadata[key] = value

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Uploading window:')
                logger.debug(MessageToDict(window))

            graphsignal._get_uploader().upload_window(window)
            if flush:
                graphsignal._get_uploader().flush_in_thread()
        else:
            with self._update_lock:
                self._current_window_updater = WindowUpdater()
                self._current_window_timestamp = time_bucket

        return True

    def log_metadata(self, key=None, value=None):
        if not isinstance(key, str) or len(key) > 250:
            logger.error('Invalid metadata entry key format')
            return
        if not isinstance(value, (str, int, float)) or len(str(value)) > 2500:
            logger.error(
                'Invalid metadata entry value format for key: {0}'.format(key))
            return

        if len(self._metadata) >= MAX_METADATA_SIZE:
            logger.error(
                'Too many metadata entries, max={0}'.format(MAX_METADATA_SIZE))
            return

        self._metadata[key] = str(value)

    def log_prediction(
            self,
            features=None,
            output=None,
            actual_timestamp=None):
        if not isinstance(features, dict):
            logger.error(
                'features (dict) must be provided')
            return

        if not isinstance(output, (bool, int, float, str, tuple, list, dict)):
            logger.error(
                'output (bool, int, float, str, tuple, list, dict) must be provided')
            return

        if isinstance(output, tuple):
            output = list(output)

        if isinstance(output, (bool, int, float, str)):
            output = [output]

        if isinstance(output, dict):
            output = list(output.values())

        self.log_prediction_batch(
            features=[list(features.values())
                      ] if features is not None else None,
            feature_names=list(
                features.keys()) if features is not None else None,
            outputs=[output] if output is not None else None,
            output_names=list(
                output.keys()) if isinstance(output, dict) else None,
            actual_timestamp=actual_timestamp
        )

    def log_prediction_batch(
            self,
            features=None,
            feature_names=None,
            outputs=None,
            output_names=None,
            actual_timestamp=None):
        if actual_timestamp is not None and not _check_timestamp(
                actual_timestamp):
            logger.error(
                'actual_timestamp (int) must be a unix timestamp in seconds')
            return

        batch_size = max(
            statistics.estimate_size(features),
            statistics.estimate_size(outputs))
        if batch_size == 0:
            logger.debug('Logged empty data')
            return

        timestamp = actual_timestamp if actual_timestamp else _timestamp()

        if not self._tumble_window(timestamp):
            return

        self._current_window_updater.add_prediction(PredictionRecord(
            features=features,
            feature_names=feature_names,
            outputs=outputs,
            output_names=output_names,
            timestamp=timestamp), batch_size)

    def log_exception(
            self,
            message=None,
            extra_info=None,
            exc_info=None,
            actual_timestamp=None):
        if not message and not exc_info:
            logger.error('Eigher of message or exc_info must be provided')
            return

        if message:
            if not isinstance(message, str):
                message = repr(message)
            if len(message) > 250:
                message = message[:250] + '...'

        stack_trace = None
        if exc_info:
            if exc_info == True:
                exc_info = sys.exc_info()
            if len(
                    exc_info) == 3 and exc_info[0] and exc_info[1] and exc_info[2]:
                exception_part = traceback.format_exception_only(
                    exc_info[0], exc_info[1])
                if len(exception_part) > 0:
                    message = str(exception_part[0]).rstrip()
                stack_trace_part = traceback.format_tb(exc_info[2])
                if len(stack_trace_part) > 0:
                    stack_trace = ''.join(stack_trace_part)

        if not message:
            logger.error('Cannot extract exception message')
            return

        if actual_timestamp is not None and not _check_timestamp(
                actual_timestamp):
            logger.error(
                'actual_timestamp (int) must be a unix timestamp in seconds')
            return

        filtered_extra_info = {}
        if extra_info and isinstance(extra_info, dict):
            for name in list(extra_info.keys())[:MAX_EXTRA_INFO_SIZE]:
                value = extra_info[name]
                name = str(name)
                if len(name) > 250:
                    name = name[:250] + '...'
                value = str(value)
                if len(value) > 2500:
                    value = value[:2500] + '...'
                filtered_extra_info[name] = value

        timestamp = actual_timestamp if actual_timestamp else _timestamp()

        if not self._tumble_window(timestamp):
            return

        self._current_window_updater.add_exception(ExceptionRecord(
            message=message,
            extra_info=filtered_extra_info if extra_info else None,
            stack_trace=stack_trace,
            timestamp=timestamp))

    def log_ground_truth(
            self,
            label,
            prediction,
            prediction_timestamp=None,
            segments=None):

        if label is None or not isinstance(
                label, (str, bool, int, float)):
            logger.error('label (str|bool|int|float) must be provided')
            return

        if prediction is None or not isinstance(
                prediction, (str, bool, int, float)):
            logger.error(
                'prediction (str|bool|int|float) must be provided')
            return

        if not isinstance(label, type(prediction)):
            logger.error(
                'label and prediction must have the same type (str|bool|int|float)')
            return

        if prediction_timestamp is not None and not _check_timestamp(
                prediction_timestamp):
            logger.error(
                'prediction_timestamp (int) must be a unix timestamp in seconds')
            return

        if segments is not None and not isinstance(segments, list):
            logger.error('segments (list) must be provided')
            return

        segments = [segment[:MAX_SEGMENT_LENGTH] for segment in segments
                    if isinstance(segment, str)]
        timestamp = prediction_timestamp if prediction_timestamp else _timestamp()

        if not self._tumble_window(timestamp):
            return

        self._current_window_updater.add_ground_truth(GroundTruthRecord(
            label=label,
            prediction=prediction,
            prediction_timestamp=timestamp,
            segments=segments))


def get_session(deployment_name):
    if not deployment_name or len(deployment_name) > 250:
        raise ValueError('Invalid deployment_name format')

    with _session_index_lock:
        if deployment_name in _session_index:
            return _session_index[deployment_name]
        else:
            sess = Session(deployment_name)
            _session_index[deployment_name] = sess
            return sess


def reset_all():
    with _session_index_lock:
        _session_index.clear()


def upload_all(force=False):
    session_list = None
    with _session_index_lock:
        session_list = _session_index.values()

    for session in session_list:
        session._tumble_window(_timestamp(), force=force, flush=False)


def _check_timestamp(timestamp):
    now = _timestamp()
    return (isinstance(timestamp, int) and
            timestamp > now - 31536000 and  # one year in the past
            timestamp < now + 86400)  # one day in the future


def _timestamp():
    return int(time.time())



import os
import time
import json
import gzip
import sys
import threading
import base64
import logging
from io import BytesIO
from urllib.request import urlopen
from urllib.request import Request
from urllib.parse import urlencode
from urllib.error import URLError
from urllib.error import HTTPError

import graphsignal
from graphsignal import metrics_pb2

logger = logging.getLogger('graphsignal')


class Uploader(object):
    MAX_BUFFER_SIZE = 2500

    def __init__(self):
        self.collector_url = 'https://log-api.graphsignal.com'
        self.buffer = []
        self.buffer_lock = threading.Lock()
        self.flush_lock = threading.Lock()

    def configure(self):
        if 'GRAPHSIGNAL_LOG_API_URL' in os.environ:
            self.collector_url = os.environ['GRAPHSIGNAL_LOG_API_URL']

    def clear(self):
        with self.buffer_lock:
            self.buffer = []

    def upload_window(self, window):
        with self.buffer_lock:
            self.buffer.append(window)
            if len(self.buffer) > self.MAX_BUFFER_SIZE:
                self.buffer = self.buffer[-self.MAX_BUFFER_SIZE:]

    def flush_in_thread(self):
        threading.Thread(target=self.flush).start()

    def flush(self):
        with self.flush_lock:
            with self.buffer_lock:
                if len(self.buffer) == 0:
                    return
                outgoing = self.buffer
                self.buffer = []
            try:
                upload_start = time.time()
                upload_request = metrics_pb2.UploadRequest()
                upload_request.windows.extend(outgoing)
                upload_request.upload_ts = _timestamp()
                resp = self._post(
                    'metrics', upload_request.SerializeToString())
                upload_response = metrics_pb2.UploadResponse()
                upload_response.ParseFromString(resp)
                logger.debug(
                    'Upload took %.3f sec', time.time() - upload_start)
            except URLError:
                logger.debug(
                    'Failed uploading metrics, will retry', exc_info=True)
                with self.buffer_lock:
                    self.buffer[:0] = outgoing
            except Exception:
                logger.error(
                    'Error uploading metrics', exc_info=True)

    def _post(self, endpoint, data):
        logger.debug('Posting data to %s/%s',
                     self.collector_url, endpoint)

        api_key_64 = _base64_encode(
            graphsignal._get_config().api_key + ':').replace('\n', '')
        headers = {
            'Accept-Encoding': 'gzip',
            'Authorization': "Basic %s" % api_key_64,
            'Content-Type': 'application/octet-stream',
            'Content-Encoding': 'gzip'
        }

        gzip_out = BytesIO()
        with gzip.GzipFile(fileobj=gzip_out, mode="w") as out_file:
            out_file.write(data)
            out_file.close()

        gzip_out_val = gzip_out.getvalue()
        if isinstance(gzip_out_val, str):
            data_gzip = bytearray(gzip_out.getvalue())
        else:
            data_gzip = gzip_out.getvalue()
        request = Request(
            url=self.collector_url + '/' + endpoint,
            data=data_gzip,
            headers=headers)

        try:
            resp = urlopen(request, timeout=10)
            result_data = resp.read()
            if resp.info():
                content_type = resp.info().get('Content-Encoding')
                if content_type == 'gzip':
                    result_data = gzip.GzipFile(
                        '', 'r', 0, BytesIO(result_data)).read()
            resp.close()
            return result_data
        except HTTPError as herr:
            logger.debug('Error message from server: %s',
                         herr.read().decode('utf-8'))
            raise herr


def _timestamp():
    return int(time.time())


def _base64_encode(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')



import time
import sys
import os
import logging
import threading
import uuid
import atexit

from graphsignal import version
from graphsignal import sessions
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')

_config = None
_uploader = None


class Config(object):
    def __init__(self):
        self.api_key = None
        self.debug_mode = None
        self.window_seconds = None
        self.buffer_size = None


def _get_config():
    return _config


def _get_uploader():
    return _uploader


def configure(api_key, debug_mode=False, window_seconds=300, buffer_size=100):
    global _config, _uploader

    if _config:
        logger.warning('Logger already configured')
        return

    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)

    if not api_key:
        logger.error('Missing argument: api_key')
        return

    _config = Config()
    _config.api_key = api_key
    _config.debug_mode = debug_mode
    _config.window_seconds = window_seconds
    _config.buffer_size = buffer_size

    _uploader = Uploader()
    _uploader.configure()

    sessions.reset_all()

    atexit.register(shutdown)

    logger.debug('Logger configured')


def shutdown():
    global _config

    if not _config:
        logger.error(
            'Logger not configured, please use graphsignal.configure()')
        return

    atexit.unregister(shutdown)

    sessions.upload_all(force=True)
    _uploader.flush()

    _config = None

    logger.debug('Logger shutdown')


def tick():
    if not _config:
        logger.error(
            'Logger not configured, please use graphsignal.configure()')
        return

    sessions.upload_all()
    _uploader.flush()


def session(deployment_name):
    if not _config:
        raise ValueError(
            'graphsignal not configured, please use graphsignal.configure()')

    return sessions.get_session(deployment_name)



import time
import unittest
import logging
from unittest.mock import patch, Mock
import sys
import numpy as np
import pandas as pd
import pprint
from google.protobuf.json_format import MessageToDict

import graphsignal
from graphsignal import statistics
from graphsignal.windows import *
from graphsignal import metrics_pb2

logger = logging.getLogger('graphsignal')


class StatisticsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(api_key='k1', debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_data_size(self):
        self.assertEqual(statistics.estimate_size([[1], [2]]), 2)
        self.assertEqual(statistics.estimate_size(np.array([[1], [2]])), 2)
        self.assertEqual(statistics.estimate_size(
            {'a': [1, 2, 3], 'b': [4, 5, 6]}), 3)
        self.assertEqual(statistics.estimate_size(
            {'a': np.asarray([1, 2, 3]), 'b': np.asarray([4, 5, 6])}), 3)
        self.assertEqual(
            statistics.estimate_size(
                pd.DataFrame(
                    data=[
                        [1],
                        [2]])),
            2)

    def test_convert_to_numpy_dict_of_lists(self):
        d1 = {'c1': [1, 2], 'c2': [3.1, 4.1]}
        d2 = {'c1': [3], 'c2': [5.1]}
        columns, data = statistics._convert_to_numpy([d1, d2])
        self.assertIsNotNone(data)
        self.assertEqual(columns, ['c1', 'c2'])
        self.assertTrue(np.array_equal(
            data[0], np.asarray([1, 2, 3])))
        self.assertTrue(np.array_equal(
            data[1], np.asarray([3.1, 4.1, 5.1])))
        self.assertEqual(data[0].dtype, 'int64')
        self.assertEqual(data[1].dtype, 'float64')

    def test_convert_to_numpy_list_of_list(self):
        d1 = [[1, 2], [3.1, 4.1]]
        d2 = [[3, 4], [5.1, 6.1]]
        columns, data = statistics._convert_to_numpy(
            [d1, d2], columns=['A', 'B'])
        self.assertIsNotNone(data)
        self.assertEqual(columns, ['A', 'B'])
        self.assertTrue(np.array_equal(
            data[0], np.asarray([1, 3.1, 3, 5.1])))
        self.assertTrue(np.array_equal(
            data[1], np.asarray([2, 4.1, 4, 6.1])))

    def test_convert_to_numpy_dataframe_with_columns(self):
        d1 = pd.DataFrame(data={'c1': [1, 2], 'c2': [3.1, 4.1]})
        d2 = pd.DataFrame(data={'c1': [3], 'c2': [5.1]})
        columns, data = statistics._convert_to_numpy([d1, d2])
        self.assertIsNotNone(data)
        self.assertEqual(columns, ['c1', 'c2'])
        self.assertTrue(np.array_equal(
            data[0], np.asarray([1, 2, 3])))
        self.assertTrue(np.array_equal(
            data[1], np.asarray([3.1, 4.1, 5.1])))
        self.assertEqual(data[0].dtype, 'int64')
        self.assertEqual(data[1].dtype, 'float64')

    def test_convert_to_numpy_dataframe_wo_columns(self):
        d = pd.DataFrame(data=[[1, 3.1], [2, 4.1]])
        columns, data = statistics._convert_to_numpy([d])
        self.assertIsNotNone(data)
        self.assertEqual(columns, ['0', '1'])
        self.assertTrue(np.array_equal(
            data[0], np.asarray([1, 2])))
        self.assertTrue(np.array_equal(
            data[1], np.asarray([3.1, 4.1])))

    def test_convert_to_numpy_ndarray(self):
        d1 = np.asarray([[1, 3.1], [2, 4.1]])
        d2 = np.asarray([[8, 12]])
        columns, data = statistics._convert_to_numpy([d1, d2])
        self.assertIsNotNone(data)
        self.assertEqual(columns, ['0', '1'])
        self.assertTrue(np.array_equal(
            data[0], np.asarray([1, 2, 8])))
        self.assertTrue(np.array_equal(
            data[1], np.asarray([3.1, 4.1, 12])))
        self.assertEqual(data[0].dtype, 'float64')
        self.assertEqual(data[1].dtype, 'float64')

    def test_update_data_metrics_empty(self):
        window = metrics_pb2.PredictionWindow()
        statistics.update_data_metrics({}, window, [])

        self.assertEqual(len(window.data_streams[str(
            metrics_pb2.DataStream.DataSource.INPUT_FEATURES)].metrics), 0)
        self.assertEqual(len(window.data_streams[str(
            metrics_pb2.DataStream.DataSource.OUTPUTS)].metrics), 0)

    def test_update_data_metrics(self):
        window = metrics_pb2.PredictionWindow()
        metric_updaters = {}

        features = {
            'f1': [1, 1, 2, 0],
            'f2': [3.5, 4.5, 5.5, 0],
            'f3': ['a', 'b', 'c', 'c'],
            'f4': [0, float('nan'), float('inf'), 2],
            'f5': [True, True, False, None]}
        outputs = [[0.1], [0.2], [0.1], [0.4]]

        statistics.update_data_metrics(
            metric_updaters, window, [
                PredictionRecord(
                    features=features, outputs=outputs)])

        for metric_updater in metric_updaters.values():
            metric_updater.finalize()

        window_dict = MessageToDict(window)

        input_metrics_json = window_dict['dataStreams'][str(
            metrics_pb2.DataStream.DataSource.INPUT_FEATURES)]['metrics']
        output_metrics_json = window_dict['dataStreams'][str(
            metrics_pb2.DataStream.DataSource.OUTPUTS)]['metrics']

        for metric in input_metrics_json.values():
            if 'distributionValue' in metric:
                del metric['distributionValue']['sketchKll10']

        for metric in output_metrics_json.values():
            if 'distributionValue' in metric:
                del metric['distributionValue']['sketchKll10']

        #pp = pprint.PrettyPrinter()
        # pp.pprint(input_metrics_json)
        # pp.pprint(output_metrics_json)

        self.assertEqual(input_metrics_json,
                         {'02ada935d60a': {'dimensions': {'feature_name': 'f5'},
                                           'distributionValue': {'sketchImpl': 'KLL10'},
                                           'name': 'distribution',
                                           'type': 'DISTRIBUTION'},
                             '076aa29f76cf': {'dimensions': {'feature_name': 'f4'},
                                              'name': 'integer_values',
                                              'ratioValue': {'counter': 2.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '222d0b1dfa42': {'dimensions': {'feature_name': 'f2'},
                                              'name': 'missing_values',
                                              'ratioValue': {'total': 4.0},
                                              'type': 'RATIO'},
                             '23777f8d6c5b': {'dimensions': {'feature_name': 'f3'},
                                              'name': 'string_values',
                                              'ratioValue': {'counter': 4.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '2df60f25840a': {'dimensions': {'feature_name': 'f3'},
                                              'distributionValue': {'sketchImpl': 'KLL10'},
                                              'name': 'distribution',
                                              'type': 'DISTRIBUTION'},
                             '31cdc4047da9': {'dimensions': {'feature_name': 'f5'},
                                              'name': 'missing_values',
                                              'ratioValue': {'counter': 1.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '342246abc789': {'dimensions': {'feature_name': 'f1'},
                                              'name': 'missing_values',
                                              'ratioValue': {'total': 4.0},
                                              'type': 'RATIO'},
                             '38b5e7971c3f': {'dimensions': {'feature_name': 'f1'},
                                              'distributionValue': {'sketchImpl': 'KLL10'},
                                              'name': 'distribution',
                                              'type': 'DISTRIBUTION'},
                             '3aa8ce8a745e': {'counterValue': {'counter': 4.0},
                                              'name': 'instance_count',
                                              'type': 'COUNTER'},
                             '4283d6f1fd2b': {'dimensions': {'feature_name': 'f5'},
                                              'name': 'boolean_values',
                                              'ratioValue': {'counter': 3.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '4a9116752484': {'dimensions': {'feature_name': 'f4'},
                                              'name': 'float_values',
                                              'ratioValue': {'total': 4.0},
                                              'type': 'RATIO'},
                             '5f5ec237ea4f': {'dimensions': {'feature_name': 'f1'},
                                              'name': 'integer_values',
                                              'ratioValue': {'counter': 4.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '62480aeb7da2': {'dimensions': {'feature_name': 'f1'},
                                              'name': 'float_values',
                                              'ratioValue': {'total': 4.0},
                                              'type': 'RATIO'},
                             '6add4609813c': {'gaugeValue': {'gauge': 5.0},
                                              'name': 'column_count',
                                              'type': 'GAUGE'},
                             '709339fc41ee': {'dimensions': {'feature_name': 'f1'},
                                              'name': 'zero_values',
                                              'ratioValue': {'counter': 1.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '72dc8927bd35': {'dimensions': {'feature_name': 'f2'},
                                              'name': 'zero_values',
                                              'ratioValue': {'counter': 1.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             '8bcee9cf32e3': {'dimensions': {'feature_name': 'f4'},
                                              'distributionValue': {'sketchImpl': 'KLL10'},
                                              'name': 'distribution',
                                              'type': 'DISTRIBUTION'},
                             'c236202ee147': {'dimensions': {'feature_name': 'f4'},
                                              'name': 'zero_values',
                                              'ratioValue': {'counter': 1.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             'd00eadf9e4cf': {'dimensions': {'feature_name': 'f2'},
                                              'name': 'integer_values',
                                              'ratioValue': {'counter': 1.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             'd082b0be0bf4': {'dimensions': {'feature_name': 'f3'},
                                              'name': 'missing_values',
                                              'ratioValue': {'total': 4.0},
                                              'type': 'RATIO'},
                             'df1ea8f7032a': {'dimensions': {'feature_name': 'f4'},
                                              'name': 'missing_values',
                                              'ratioValue': {'counter': 2.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             'e2844d940cc2': {'dimensions': {'feature_name': 'f2'},
                                              'name': 'float_values',
                                              'ratioValue': {'counter': 3.0, 'total': 4.0},
                                              'type': 'RATIO'},
                             'e675c4eab4f7': {'dimensions': {'feature_name': 'f3'},
                                              'name': 'empty_values',
                                              'ratioValue': {'total': 4.0},
                                              'type': 'RATIO'},
                             'f1d8c84f680a': {'dimensions': {'feature_name': 'f2'},
                                              'distributionValue': {'sketchImpl': 'KLL10'},
                                              'name': 'distribution',
                                              'type': 'DISTRIBUTION'}})
        self.assertEqual(output_metrics_json,
                         {'094ce9a486c0': {'gaugeValue': {'gauge': 1.0},
                                           'name': 'column_count',
                                           'type': 'GAUGE'},
                          '15f3585b5a2b': {'dimensions': {'output_name': '0'},
                                           'name': 'missing_values',
                                           'ratioValue': {'total': 4.0},
                                           'type': 'RATIO'},
                          '2d4c320b727b': {'dimensions': {'output_name': '0'},
                                           'name': 'integer_values',
                                           'ratioValue': {'total': 4.0},
                                           'type': 'RATIO'},
                          '350dc33d58cf': {'dimensions': {'output_name': '0'},
                                           'name': 'zero_values',
                                           'ratioValue': {'total': 4.0},
                                           'type': 'RATIO'},
                          '3dea0616b753': {'dimensions': {'output_name': '0'},
                                           'name': 'float_values',
                                           'ratioValue': {'counter': 4.0, 'total': 4.0},
                                           'type': 'RATIO'},
                          '59abb90a51e4': {'dimensions': {'output_name': '0'},
                                           'distributionValue': {'sketchImpl': 'KLL10'},
                                           'name': 'distribution',
                                           'type': 'DISTRIBUTION'},
                          '8e1bc8013472': {'counterValue': {'counter': 4.0},
                                           'name': 'instance_count',
                                           'type': 'COUNTER'}})

    def test_update_data_metrics_perf(self):
        window = metrics_pb2.PredictionWindow()
        metric_updaters = {}

        prediction_buffer = []
        for i in range(100):
            features = {}
            for j in range(50):
                features['f1_' + str(j)] = [i]
                features['f2_' + str(j)] = [i + 3.53242342]
                features['f3_' + str(j)] = [100 * 'abc' + str(i)]
                features['f4_' + str(j)] = [float('nan')]
                features['f4_' + str(j)] = [float('+inf')]
            outputs = [[0.1], [0.2], [0.1], [0.4]]
            prediction_buffer.append(
                PredictionRecord(
                    features=features,
                    outputs=outputs))

        #import cProfile
        #from pstats import Stats, SortKey

        start_ts = time.time()

        # with cProfile.Profile() as pr:
        for k in range(10):
            statistics.update_data_metrics(
                metric_updaters, window, prediction_buffer)
        #stats = Stats(pr)
        # stats.sort_stats(SortKey.CUMULATIVE).print_stats(25)

        for metric_updater in metric_updaters.values():
            metric_updater.finalize()

        took = time.time() - start_ts

        print('update_data_metrics took (sec): ', round(took, 6))
        self.assertTrue(len(window.data_streams[str(
            metrics_pb2.DataStream.DataSource.INPUT_FEATURES)].metrics) > 0)
        self.assertTrue(len(window.data_streams[str(
            metrics_pb2.DataStream.DataSource.OUTPUTS)].metrics) > 0)
        self.assertTrue(took < 1)

    def test_truncate_strings(self):
        truncated = statistics._truncate_strings(
            ['abc', '1234567890abcdefgh', '1234567890abcdefghi'])
        self.assertEqual(
            truncated, [
                'abc', '1234567890abcdefgh', '1234567890...efghi'])

    def test_update_performance_metrics_binary(self):
        window = metrics_pb2.PredictionWindow()
        metric_updaters = {}

        ground_truth_buffer = [
            GroundTruthRecord(
                label=True,
                prediction=True,
                segments=[
                    's1',
                    's2']),
            GroundTruthRecord(
                label=True,
                prediction=True,
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label=True,
                prediction=False,
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label=False,
                prediction=True,
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label=False,
                prediction=False,
                segments=[
                    's1',
                    's2'])
        ]

        statistics.update_performance_metrics(
            metric_updaters, window, ground_truth_buffer)

        for metric_updater in metric_updaters.values():
            metric_updater.finalize()

        window_dict = MessageToDict(window)
        metrics_json = window_dict['dataStreams'][str(
            metrics_pb2.DataStream.DataSource.GROUND_TRUTH_BINARY)]['metrics']

        #pp = pprint.PrettyPrinter()
        # pp.pprint(metrics_json)

        self.assertEqual(metrics_json, {'2a8a320fc710': {'name': 'accuracy',
                                                         'ratioValue': {'counter': 3.0, 'total': 5.0},
                                                         'type': 'RATIO'},
                                        '2a915536a163': {'counterValue': {'counter': 1.0},
                                                         'name': 'binary_false_positives',
                                                         'type': 'COUNTER'},
                                        '3e98594499f3': {'counterValue': {'counter': 1.0},
                                                         'name': 'binary_false_negatives',
                                                         'type': 'COUNTER'},
                                        '4769fae29cea': {'counterValue': {'counter': 1.0},
                                                         'name': 'binary_true_negatives',
                                                         'type': 'COUNTER'},
                                        'a335baad03d4': {'distributionValue': {'sketchImpl': 'KLL10',
                                                                               'sketchKll10': {'H': '1',
                                                                                               'c': 0.6666666666666666,
                                                                                               'compactorsString': [{'items': ['s1',
                                                                                                                               's2',
                                                                                                                               's1',
                                                                                                                               's3',
                                                                                                                               's1',
                                                                                                                               's2']}],
                                                                                               'itemType': 'STRING',
                                                                                               'k': '10',
                                                                                               'maxSize': '11',
                                                                                               'size': '6'}},
                                                         'name': 'segment_matches',
                                                         'type': 'DISTRIBUTION'},
                                        'e98d593706c6': {'distributionValue': {'sketchImpl': 'KLL10',
                                                                               'sketchKll10': {'H': '1',
                                                                                               'c': 0.6666666666666666,
                                                                                               'compactorsString': [{'items': ['s1',
                                                                                                                               's2',
                                                                                                                               's1',
                                                                                                                               's3',
                                                                                                                               's1',
                                                                                                                               's3',
                                                                                                                               's1',
                                                                                                                               's3',
                                                                                                                               's1',
                                                                                                                               's2']}],
                                                                                               'itemType': 'STRING',
                                                                                               'k': '10',
                                                                                               'maxSize': '11',
                                                                                               'size': '10'}},
                                                         'name': 'segment_totals',
                                                         'type': 'DISTRIBUTION'},
                                        'ee427ea584e7': {'counterValue': {'counter': 2.0},
                                                         'name': 'binary_true_positives',
                                                         'type': 'COUNTER'}})

    def test_update_performance_metrics_categorical(self):
        window = metrics_pb2.PredictionWindow()
        metric_updaters = {}

        ground_truth_buffer = [
            GroundTruthRecord(
                label='c1',
                prediction='c1',
                segments=[
                    's1',
                    's2']),
            GroundTruthRecord(
                label='c1',
                prediction='c1',
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label='c1',
                prediction='c2',
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label='c1',
                prediction='c2',
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label='c2',
                prediction='c1',
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label='c2',
                prediction='c2',
                segments=[
                    's1',
                    's2']),
            GroundTruthRecord(
                label='c3',
                prediction='c3',
                segments=[
                    's1',
                    's2']),
        ]

        statistics.update_performance_metrics(
            metric_updaters, window, ground_truth_buffer)

        for metric_updater in metric_updaters.values():
            metric_updater.finalize()

        window_dict = MessageToDict(window)
        metrics_json = window_dict['dataStreams'][str(
            metrics_pb2.DataStream.DataSource.GROUND_TRUTH_CATEGORICAL)]['metrics']

        data_stream = window.data_streams[str(
            metrics_pb2.DataStream.DataSource.GROUND_TRUTH_CATEGORICAL)]

        total = None
        sketches = {}
        for metric in data_stream.metrics.values():
            if metric.name == 'total':
                total = metric.counter_value.counter
            elif metric.name == 'class_true_positives':
                sketches['class_true_positives'] = KLLSketch()
                sketches['class_true_positives'].from_proto(
                    metric.distribution_value.sketch_kll10)
            elif metric.name == 'class_false_positives':
                sketches['class_false_positives'] = KLLSketch()
                sketches['class_false_positives'].from_proto(
                    metric.distribution_value.sketch_kll10)
            elif metric.name == 'class_false_negatives':
                sketches['class_false_negatives'] = KLLSketch()
                sketches['class_false_negatives'].from_proto(
                    metric.distribution_value.sketch_kll10)

        # c1 2f22765d
        # c2 6b1f5330
        # c3 a625406f
        self.assertEqual(total, 7.0)
        self.assertEqual(sketches['class_true_positives'].distribution(),
                         [['2f22765d', 2], ['6b1f5330', 1], ['a625406f', 1]])
        self.assertEqual(sketches['class_false_positives'].distribution(),
                         [['2f22765d', 1], ['6b1f5330', 2]])
        self.assertEqual(sketches['class_false_negatives'].distribution(),
                         [['2f22765d', 2], ['6b1f5330', 1]])

    def test_update_performance_metrics_numeric(self):
        window = metrics_pb2.PredictionWindow()
        metric_updaters = {}

        ground_truth_buffer = [
            GroundTruthRecord(
                label=1.0,
                prediction=3.4,
                segments=[
                    's1',
                    's2']),
            GroundTruthRecord(
                label=2.4,
                prediction=0.2,
                segments=[
                    's1',
                    's3'])
        ]

        statistics.update_performance_metrics(
            metric_updaters, window, ground_truth_buffer)

        for metric_updater in metric_updaters.values():
            metric_updater.finalize()

        window_dict = MessageToDict(window)
        metrics_json = window_dict['dataStreams'][str(
            metrics_pb2.DataStream.DataSource.GROUND_TRUTH_NUMERIC)]['metrics']

        #pp = pprint.PrettyPrinter()
        # pp.pprint(metrics_json)

        self.assertEqual(metrics_json, {'2789175b4787': {'counterValue': {'counter': 4.6},
                                                         'name': 'mae_sum',
                                                         'type': 'COUNTER'},
                                        '8c82f7b3ed54': {'counterValue': {'counter': 2.0},
                                                         'name': 'mse_n',
                                                         'type': 'COUNTER'},
                                        'fe4a651921fe': {'counterValue': {'counter': 10.599999999999998},
                                                         'name': 'mse_sum',
                                                         'type': 'COUNTER'}})

    def test_update_performance_metrics_segments(self):
        window = metrics_pb2.PredictionWindow()
        metric_updaters = {}

        ground_truth_buffer = [
            GroundTruthRecord(
                label=True,
                prediction=True,
                segments=[
                    's1',
                    's2']),
            GroundTruthRecord(
                label=True,
                prediction=True,
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label=True,
                prediction=False,
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label=False,
                prediction=True,
                segments=[
                    's1',
                    's3']),
            GroundTruthRecord(
                label=False,
                prediction=False,
                segments=[
                    's1',
                    's2'])
        ]

        statistics.update_performance_metrics(
            metric_updaters, window, ground_truth_buffer)

        for metric_updater in metric_updaters.values():
            metric_updater.finalize()

        data_stream = window.data_streams[str(
            metrics_pb2.DataStream.DataSource.GROUND_TRUTH_BINARY)]

        sketches = {}
        for metric in data_stream.metrics.values():
            if metric.name == 'segment_totals':
                sketches['segment_totals'] = KLLSketch()
                sketches['segment_totals'].from_proto(
                    metric.distribution_value.sketch_kll10)
            elif metric.name == 'segment_matches':
                sketches['segment_matches'] = KLLSketch()
                sketches['segment_matches'].from_proto(
                    metric.distribution_value.sketch_kll10)

        self.assertEqual(sketches['segment_totals'].distribution(),
                         [['s1', 5], ['s2', 2], ['s3', 3]])
        self.assertEqual(sketches['segment_matches'].distribution(), [
                         ['s1', 3], ['s2', 2], ['s3', 1]])
